#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class PlaybookTemplatesController < ApplicationController
  helper Redborder::ModulesHelper
  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_playbook_template, only: [:edit, :update, :destroy, :show, :create_clone, :set_as_default]
  before_action :check_privileges, only: [:create, :create_clone, :set_as_default, :update, :destroy,
                                          :create_task_template, :update_task_template, :destroy_task_template,
                                          :create_automatic_action_template, :edit_automatic_action_template]
  before_action :set_incident_fields, only: [:new, :edit, :edit_automatic_action_template]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all playbook templates in a table
  #
  def index
    @playbook_templates = PlaybookTemplate.order(created_at: :asc)

    # Paginate results
    @page = params[:page]
    @playbook_templates = @playbook_templates.paginate(page: params[:page],
                               per_page: current_user.per_page_count)
  end

  #
  # => Render the form to create a new playbook template
  #
  def new
    @playbook_template = PlaybookTemplate.new
  end

  #
  # => Receive the data to create a new playbook template. Try to create a new playbook template
  #    and redirect to playbook_templates_path if it's ok
  #
  def create
    begin
      @playbook_template = current_user.playbook_templates.build(playbook_template_params)

      @playbook_template.save

      # Create default phase templates from the default playbook template
      # TODO: protect this
      default_playbook_template = PlaybookTemplate.first
      if default_playbook_template
        default_playbook_template.phase_templates.each do |phase_template|
          new_phase_template = phase_template.dup
          new_phase_template.playbook_template = @playbook_template
          new_phase_template.save!
        end
      end

      # Create automatic actions
      if params[:playbook_template][:automatic_actions] == '1' && !params[:automatic_actions].nil? && !params[:automatic_actions].empty?
        params[:automatic_actions].each do |integration, actions|
          actions.each do |action, action_instance|
            action_instance.each do |instance, instance_params|
              @automatic_action = @playbook_template.automatic_action_templates.create!(
                integration_service: integration,
                action: action
              )
              incident_params = {}

              unless instance_params.empty?
                instance_params.each do |param_key, param_value|
                  incident_params[param_key] = param_value
                end
              end
              @automatic_action.update(params: incident_params) unless incident_params.empty?
            end
          end
        end
      end

      respond_to do |format|
          format.html do
            redirect_to playbook_templates_path,
                        notice: "Playbook was successfully made"
          end
          format.js { flash[:notice] = "Playbook was successfully made" }
          format.json { head :no_content }
      end
    rescue => e
      Rails.logger.error "Error in create playbook template: #{e.message}"
      respond_to do |format|
        format.html { redirect_to playbook_templates_path, alert: 'Error creating playbook' }
        format.js do
            flash[:error] = @playbook_template.errors.full_messages.map do
              |msg| "<li>#{msg}</li>"
            end.join
        end
        format.json do
          render json: @playbook_template.errors, status: :unprocessable_entity
        end
      end
    end
  end

  def create_clone
    begin
      new_playbook_template = nil
      ActiveRecord::Base.transaction do
        new_playbook_template = @playbook_template.dup
        new_playbook_template.read_only = false # Important when cloning the default playbook
        new_playbook_template.default = false
        new_playbook_template.name = "#{@playbook_template.name} (Clone)"
        new_playbook_template.user_id = current_user.id
        new_playbook_template.save!

        @playbook_template.phase_templates.each do |phase_template|
          new_phase_template = phase_template.dup
          new_phase_template.playbook_template = new_playbook_template
          new_phase_template.save!

          phase_template.task_templates.each do |task_template|
            new_task_template = task_template.dup
            new_task_template.phase_template = new_phase_template
            new_task_template.save!
          end
        end

        @playbook_template.automatic_action_templates.each do |automatic_action_template|
          new_automatic_action_template = automatic_action_template.dup
          new_automatic_action_template.playbook_template = new_playbook_template
          new_automatic_action_template.save!
        end
      end

      rb_flash 'Playbook was successfully created'
    rescue => e
      rb_flash "Failed to clone playbook: #{e.message}", type: :error
    end

    redirect_to playbook_templates_path
  end

  def set_as_default
    respond_to do |format|
      begin
        @playbook_template.set_as_default
        format.html do
          notice_message = "Playbook was successfully updated"

          if request.referer && request.referer.include?(playbook_template_path(@playbook_template))
            redirect_to playbook_template_path(@playbook_template), notice: notice_message
          else
            redirect_to playbook_templates_path, notice: notice_message
          end
        end
        format.js { flash[:notice] = "Playbook was successfully updated" }
        format.json { render json: @playbook_template }
      rescue => e
        format.html { render action: 'show', alert: e.message }
        format.js do
          flash[:error] = e.message
        end
        format.json { render json: { error: e.message }, status: :unprocessable_entity }
      end
    end
  end

  #
  # => Render the form to edit a playbook
  #
  def edit
    respond_to do |format|
      format.js
    end
  end

  #
  # => Receive the data to update a new playbook_template. Try to update a new playbook_template
  #    and redirect to playbook_template_path if it's ok
  #
  def update
    respond_to do |format|
      if @playbook_template.update(playbook_template_params)
        process_automatic_actions if params[:automatic_actions]

        format.html do
          notice_message = "Playbook was successfully updated"

          if request.referer && request.referer.include?(playbook_template_path(@playbook_template))
            redirect_to playbook_template_path(@playbook_template), notice: notice_message
          else
            redirect_to playbook_template_path, notice: notice_message
          end
        end
        format.js { flash[:notice] = "Playbook was successfully updated" }
        format.json { render json: @playbook_template }
      else
        format.html { render action: 'show' }
        format.js do
          flash[:error] = @playbook_template.errors.full_messages.map do |msg|
            "<li>#{msg}</li>"
          end.join
        end
        format.json { render json: @playbook_template.errors, status: :unprocessable_entity }
      end
    end
  end

  #
  # => Destroy an playbook_template
  #
  def destroy
    @playbook_template.destroy
    respond_to do |format|
      format.html { redirect_to playbook_templates_path }
      format.js
      format.json { head :no_content }
    end
  end

  #
  # Search on incidents
  #
  def search
    @page = params[:page] || 1

    filter = ['( LOWER(name) LIKE :search or ' \
              'LOWER(description) LIKE :search )',
              search: "%#{params[:fast_search].try(:downcase)}%"]

    @playbook_templates = PlaybookTemplate.where(filter).order(created_at: :desc)
      .paginate(page: @page, per_page: current_user.per_page_count)
  end

  def show
  end

  def new_task_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @phase_template = PhaseTemplate.find(params[:phase_template_id])
    @task_template = TaskTemplate.new

    respond_to do |format|
      format.js
    end
  end

  def new_automatic_action
    @playbook_template = PlaybookTemplate.find(params[:id])
    @automatic_action = AutomaticActionTemplate.new

    respond_to do |format|
      format.js
    end
  end

  def create_task_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @phase_template = PhaseTemplate.find(task_template_params[:phase_template_id])
    @task_template = TaskTemplate.new(task_template_params)

    @task_template.phase_template = @phase_template

    @phase_template.task_templates << @task_template

    respond_to do |format|
      if @task_template.save
        format.html do
          redirect_to playbook_template_path(@playbook_template, anchor: @phase_template.name.downcase), notice: "Task was successfully created"
        end
        format.js { flash[:notice] = "Task Template was successfully created" }
        format.json { head :no_content }
      else
        format.html { redirect_to playbook_template_path(@playbook_template, anchor: @phase_template.name.downcase), alert: 'Error creating task' }
        format.js do
          flash[:error] = @task_template.errors.full_messages.join(", ")
        end
        format.json { render json: @task_template.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_automatic_action_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @automatic_action = AutomaticActionTemplate.new(automatic_action_template_params)
  end

  def edit_task_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @task_template = TaskTemplate.find(params[:task_template_id])
    @phase_template = @task_template.phase_template

    respond_to do |format|
      format.js
    end
  end

  def edit_automatic_action_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @automatic_action_templates = AutomaticActionTemplate.where(playbook_template_id: @playbook_template.id)
    respond_to do |format|
      format.js
    end
  end

  def update_task_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @task_template = TaskTemplate.find(task_template_params[:id])
    @phase_template = @task_template.phase_template

    respond_to do |format|
      if @task_template.update(task_template_params)
        format.html do
          redirect_to playbook_template_path(@playbook_template, anchor: @phase_template.name.downcase), notice: "Task was successfully created"
        end
        format.js { flash[:notice] = "Playbook was successfully updated" }
        format.json { render json: @playbook_template }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @task_template.errors.full_messages.join(", ")
        end
        format.json { render json: @task_template.errors, status: :unprocessable_entity }
      end
    end
  end

  def destroy_task_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @task_template = TaskTemplate.find(params[:task_template_id])

    @phase_template = @task_template.phase_template

    @task_template.destroy

    respond_to do |format|
      format.html do
        redirect_to playbook_template_path(@playbook_template, anchor: @phase_template.name.parameterize), notice: "Task was successfully destroyed"
      end
      format.js
      format.json { head :no_content }
    end
  end

  def integration_service_actions
    integration_service = params[:integration_service]
    @actions = IntegrationService.get_actions_by_integration(integration_service).keys rescue []

    respond_to do |format|
      format.json { render json: @actions }
    end
  end

  def integration_service_action_params
    integration_service = params[:integration_service]
    action = params[:action_name]
    @action_params = IntegrationService.get_action_params(integration_service, action) rescue {}

    respond_to do |format|
      format.json { render json: @action_params }
    end
  end

  private

  #
  # => Init the playbook template if the user has permissions
  #
  def set_playbook_template
    @playbook_template = PlaybookTemplate.find(params[:id])
    @automatic_action_templates = AutomaticActionTemplate.where(playbook_template_id: @playbook_template)
  end

  def check_privileges
    current_user.admin?
  end

  def set_incident_fields
    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }
    modules_columns = {}
    incident_modules.each do |k, v|
      name = v[:name]
      name = name.to_s.start_with?("tr.") ? I18n.t(name.slice(3, name.length - 3)) : name
      module_data = {}
      v[:columns].each do |column, column_data|
        parsed_data = {:columns => column_data[:columns], :category => column_data[:category]}
        parsed_column = {column.to_sym => parsed_data}
        module_data[column.to_sym] = parsed_data
      end
      modules_columns[name] = module_data
    end

    @modules =  [[I18n.t('controllers.playbook_assignment_rules.incidents'), 'Incidents'], ['Intrusion', 'Intrusion'], ['Vault', 'Vault']]
    @incident_categories = {'Incidents' => ['General']}
    @incident_fields = {
      'Incidents' => {'General' => Incident.column_names
                    .map { |col| {col => I18n.t("incidents.columns.#{col}")}}
    }}
    modules_columns.each do |module_name, module_data|
      @incident_categories[module_name] ||= []
      @incident_fields[module_name] ||= {}

      module_data.each do |_, column|
        @incident_categories[module_name] |= [column[:category]]
        @incident_fields[module_name][column[:category]] ||= []

        column[:columns].each do |key, name|
          translated_name = name.to_s.start_with?("tr.") ? I18n.t(name.slice(3, name.length - 3)) : name
          @incident_fields[module_name][column[:category]] |= [{key => translated_name}]
        end
      end
    end
  end

  def automatic_actions_enabled?
    params.dig(:playbook_template, :automatic_actions) == '1' && params[:automatic_actions].present?
  end

  def process_automatic_actions
    return @automatic_action_templates.destroy_all unless automatic_actions_enabled?

    desired_actions = []
    params[:automatic_actions].each do |integration, actions|
      actions.each do |action, instances|
        instances.each do |_instance_key, instance_params|
          desired_actions << {
            integration_service: integration,
            action: action,
            params: instance_params.presence || {}
          }
        end
      end
    end

    existing_actions = @automatic_action_templates.to_a
    matched_actions = []
    desired_actions.each do |desired|
      match = existing_actions.find do |existing|
        existing.integration_service == desired[:integration_service] &&
        existing.action == desired[:action]
      end

      if match
        matched_actions << match
        if match.params != desired[:params]
          match.update!(params: desired[:params])
        end
        existing_actions.delete(match)
      else
        @playbook_template.automatic_action_templates.create!(desired)
      end
    end
    existing_actions.each(&:destroy)
  end

  #
  # => Return a hash with the params to create/update a playbook template
  #
  def playbook_template_params
    params
      .require(:playbook_template)
      .permit(:name, :description, :phase_template_id, :summary, :page, :search, :fast_search, :automatic_action, :automatic_actions)
  end

  def task_template_params
    params
      .require(:task_template)
      .permit(:id, :name, :description, :summary, :phase_template_id, :page, :execution, :integration_service, :action)
  end
end