#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'yaml'

class DeviceTypeClassifier
  CONFIG_PATH = Rails.root.join('config', 'device_type_classifier.yml')

  def self.classify(device)
    return nil unless device

    config = load_config

    Delayed::Worker.logger.info "Checking classification for InventoryDevice: #{device.id}"

    # Classification by device name
    if device.name
      name = device.name.split(".").first
      Delayed::Worker.logger.info "...using name #{name} for classification for InventoryDevice: #{device.id}"
      inventory_device_type_object = classify_by_name(name, config["name_mapping"]) if name
      if inventory_device_type_object
        return inventory_device_type_object
      else
        Delayed::Worker.logger.info "...no name classification found"
      end
    else
      Delayed::Worker.logger.info "...skipping name classification (no name available)"
    end

    # Classification by MAC vendor
    device.mac_objects.map { |mac| mac[:vendor] }.compact.each do |vendor|
      Delayed::Worker.logger.info "...using vendor #{vendor} for classification for InventoryDevice: #{device.id}"
      inventory_device_type_object = classify_by_mac_vendor(vendor, config["vendor_mapping"]) if vendor
      if inventory_device_type_object
        return inventory_device_type_object
      else
        Delayed::Worker.logger.info "...no classification found by vendor"
      end
    end

    nil
  end

  private

  # Load the configuration from the YAML file
  def self.load_config
    @config ||= YAML.load_file(CONFIG_PATH)
  rescue Errno::ENOENT
    Rails.logger.error("Config file not found: #{CONFIG_PATH}")
    {}
  end

  # Classify by MAC vendor
  def self.classify_by_mac_vendor(vendor, vendor_map)
    return nil unless vendor_map

    vendor_map.each do |type_name, patterns|
      patterns.each do |pattern|
        return InventoryDeviceTypeObject.find_by(name: type_name) if vendor.match?(/#{pattern}/i)
      end
    end

    nil
  end

  # Classify by device name
  def self.classify_by_name(name, name_map)
    return nil unless name_map

    name_map.each do |type_name, patterns|
      patterns.each do |pattern|
        return InventoryDeviceTypeObject.find_by(name: type_name) if name.match?(/#{pattern}/i)
      end
    end

    nil
  end
end
